/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_dccallow.c: DCCALLOW command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_dccallow.c,v 1.14 2002/12/15 21:26:17 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_dccallow.so",
	"/DCCALLOW command",
	4,
	"$Revision: 1.14 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("DCCALLOW", m_dccallow, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("DCCALLOW", m_dccallow);
	return;
}
#endif

static int add_dccallow(aClient *sptr, aClient *acptr)
{
	Link *lp;
	int cnt = 0;

	for (lp = sptr->user->dccallow; lp != NULL; lp = lp->next) {
		if (lp->flags != DCC_LINK_ME) {
			continue;
		}
		if (++cnt >= MAXDCCALLOW) {
			sendto_one(sptr, err_str(ERR_TOOMANYDCC), me.name, sptr->name, acptr->name,
				MAXDCCALLOW);
			return 0;
		}
		else if (lp->value.cptr == acptr) {
			return 0;
		}
	}

	lp = make_slink();
	lp->value.cptr = acptr;
	lp->flags = DCC_LINK_ME;
	lp->next = sptr->user->dccallow;
	sptr->user->dccallow = lp;

	lp = make_slink();
	lp->value.cptr = sptr;
	lp->flags = DCC_LINK_REMOTE;
	lp->next = acptr->user->dccallow;
	acptr->user->dccallow = lp;

	sendto_one(sptr, rpl_str(RPL_DCCSTATUS), me.name, sptr->name, acptr->name, "added to");
	return 0;
}

static int del_dccallow(aClient *sptr, aClient *acptr)
{
	Link *lp, **list;
	int found = 0;

	for (list = &(sptr->user->dccallow); *list; list = &((*list)->next)) {
		if ((*list)->flags != DCC_LINK_ME) {
			continue;
		}
		if ((*list)->value.cptr == acptr) {
			lp = *list;
			*list = lp->next;
			free_slink(lp);
			found++;
			break;
		}
	}
	if (!found) {
		sendto_one(sptr, ":%s %d %s :%s is not in your DCC allow list", me.name, RPL_DCCINFO,
			sptr->name, acptr->name);
		return 0;
	}

	found = 0;
	for (list = &(acptr->user->dccallow); *list; list = &((*list)->next)) {
		if ((*list)->flags != DCC_LINK_REMOTE) {
			continue;
		}
		if ((*list)->value.cptr == sptr) {
			lp = *list;
			*list = lp->next;
			free_slink(lp);
			found++;
			break;
		}
	}
	if (!found) {
		ircdlog(LOG_DEFAULT, "%s was in dccallowme list for %s but nto in dccallowremote list",
			acptr->name, sptr->name);
	}
	sendto_one(sptr, rpl_str(RPL_DCCSTATUS), me.name, sptr->name, acptr->name, "removed from");
	return 0;
}

/*
 * m_dccallow
 *	parv[0] = sender prefix
 *	parv[1] = parameter
 */
int m_dccallow(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	Link *lp;
	char *p = NULL, *s, *cn, **ptr;
	aClient *acptr, *lastcptr = NULL;
	int didlist = 0, didhelp = 0, didanything = 0;
	static char *dcc_help[] = {
		"/DCCALLOW [<+|->nick[,<+|->nick, ...]] [list] [help]",
		" ",
		"You may allow DCC sends of banned filetypes which are otherwise blocked by",
		"the server using the DCCALLOW command. You simply add the user(s) you wish to",
		"accept the dcc from to your dcc allow list. For example, if you would like",
		"to receive a banned file from the user Bob, you would type the following",
		"command and Bob will be free to send you any files.",
		" ",
		"/quote dccallow +Bob",
		" ",
		"If you want to then remove a user from your dcc allow list, you can do so by",
		"using the following command. Again, in the example, user Bob is being removed.",
		" ",
		"/quote dccallow -Bob",
		" ",
		"You can also add and remove more than one user from your dcc allow list at",
		"the same time, for example:",
		" ",
		"/quote dccallow +user1,-user2,+user3",
		" ",
		"To see a list of all users on your dcc allow list, use the 'list' parameter,",
		"and to see this help text again, use the 'help' parameter.",
		NULL
	};

	if (!MyClient(sptr)) {
		return 0;
	}
	if (parc < 2) {
		sendto_one(sptr, ":%s NOTICE %s :Type /DCCALLOW HELP for more information on the DCC "
			"Allow system.", me.name, parv[0]);
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "DCCALLOW");
		return 0;
	}
	for (s = strtoken(&p, parv[1], ","); s; s = strtoken(&p, NULL, ",")) {
		while (*s == ' ') {
			s++;
		}
		if (*s == '+') {
			didanything++;
			cn = s + 1;
			if (*cn == '\0') {
				continue;
			}
			if ((acptr = find_person(cn, NULL)) == NULL) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], cn);
				continue;
			}
			if (acptr == sptr) {
				continue;
			}
			if (acptr == lastcptr) {
				sendto_realops_lev(SPY_LEV, "User %s (%s@%s) may be flooding DCCALLOW: add %s",
					sptr->name, sptr->username, MaskedHost(sptr), cn);
			}
			lastcptr = acptr;
			add_dccallow(sptr, acptr);
		}
		else if (*s == '-') {
			didanything++;
			cn = s + 1;
			if (*cn == '\0') {
				continue;
			}
			if ((acptr = find_person(cn, NULL)) == NULL) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], cn);
				continue;
			}
			if (acptr == sptr) {
				continue;
			}
			if (acptr == lastcptr) {
				sendto_realops_lev(SPY_LEV, "User %s (%s@%s) may be flooding DCCALLOW: del %s",
					sptr->name, sptr->username, MaskedHost(sptr), cn);
			}
			lastcptr = acptr;
			del_dccallow(sptr, acptr);
		}
		else {
			if (!didlist && !mycmp(s, "LIST")) {
				didanything++;
				didlist = 1;

				sendto_one(sptr, ":%s %d %s :The following users are on your DCC allow list:",
					me.name, RPL_DCCINFO, sptr->name);
				for (lp = sptr->user->dccallow; lp != NULL; lp = lp->next) {
					if (lp->flags == DCC_LINK_REMOTE) {
						continue;
					}
					acptr = lp->value.cptr;
					sendto_one(sptr, ":%s %d %s :%s (%s@%s)", me.name, RPL_DCCLIST,
						sptr->name, acptr->name, acptr->username, MaskedHost(acptr));
				}
				sendto_one(sptr, rpl_str(RPL_ENDOFDCCLIST), me.name, parv[0], s);
			}
			else if (!didhelp && !mycmp(s, "HELP")) {
				didanything++;
				didhelp = 1;
				for (ptr = dcc_help; *ptr; ptr++) {
					sendto_one(sptr, ":%s %d %s :%s", me.name, RPL_DCCINFO, parv[0], *ptr);
				}
				sendto_one(sptr, rpl_str(RPL_ENDOFDCCLIST), me.name, parv[0], s);
			}
		}
	}
	if (!didanything) {
		sendto_one(sptr, ":%s NOTICE %s :Type /DCCALLOW HELP for more information on the DCC allow system.",
			me.name, parv[0]);
	}
	return 0;
}
