/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_kill.c: KILL command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_kill.c,v 1.12 2002/12/15 21:26:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static char buf[BUFSIZE];

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_kill.so",
	"/KILL command",
	4,
	"$Revision: 1.12 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("KILL", m_kill, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("KILL", m_kill);
	return;
}
#endif

/*
 * m_kill
 *	parv[0] = sender prefix
 *	parv[1] = target
 *	parv[2] = path
 */
int m_kill(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr;
	char *user, *path, *killer, *p, *nick;
	char mypath[KILLLEN + 1];
	int chasing = 0, kcount = 0;

	if (!IsPrivileged(cptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}
	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KILL");
		return 0;
	}
	user = parv[1];
	if ((path = parv[2]) == NULL) {
		path = ")";
	}
	if (!BadPtr(path)) {
		if (strlen(path) > KILLLEN) {
			path[KILLLEN] = '\0';
		}
	}
	if (MyClient(sptr)) {
		user = canonize(user);
	}
	for (p = NULL, nick = strtoken(&p, user, ","); nick != NULL; nick = strtoken(&p, NULL, ",")) {
		chasing = 0;
		if ((acptr = find_client(nick, NULL)) == NULL) {
			if ((acptr = get_history(nick, (long)KILLCHASETIMELIMIT)) == NULL) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
				continue;
			}
			sendto_one(sptr, ":%s NOTICE %s :KILL target changed from %s to %s", me.name,
				parv[0], nick, acptr->name);
			chasing = 1;
		}

		if (MyClient(cptr) && ((MyConnect(acptr) && !OPHasFlag(cptr, OFLAG_LKILL)) ||
		  (!MyConnect(acptr) && !OPHasFlag(cptr, OFLAG_GKILL)))) {
			sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
			continue;
		}
		if (IsServer(acptr) || IsMe(acptr) || (HasMode(acptr, UMODE_SCLIENT) && !IsServer(cptr))) {
			sendto_one(sptr, err_str(ERR_CANTKILLSERVER), me.name, parv[0]);
			ircdlog(LOG_KILL, "client %s attempted to kill server %s",
				get_client_name(cptr, FALSE), acptr->name);
			continue;
		}

		kcount++;
		if (!IsServer(sptr) && kcount > MAXKILLS) {
			sendto_one(sptr, ":%s NOTICE %s :Too many targets, kill list truncated (maximum is %d)",
				me.name, parv[0], MAXKILLS);
			break;
		}
		if (MyClient(sptr)) {
			char myname[HOSTLEN + 1], *s;
			int slen;

			strncpy(myname, me.name, HOSTLEN + 1);
			if ((s = index(myname, '.')) != NULL) {
				*s = '\0';
			}

			if ((slen = KILLLEN - (strlen(myname) + strlen(MaskedHost(sptr)) +
			  strlen(sptr->name) + 8)) < 0) {
				slen = 0;
			}
			if (strlen(path) > slen) {
				path[slen] = '\0';
			}
			ircsprintf(mypath, "%s!%s!%s (%s)", myname, MaskedHost(sptr), sptr->name, path);
		}
		else {
			strncpy(mypath, path, KILLLEN + 1);
		}

		sendto_ops_lev(HasMode(sptr, UMODE_OPER) ? 0 : SKILL_LEV, "Received KILL message for %s!%s@%s. "
			"From: %s Path: %s", acptr->name, acptr->user ? acptr->username : "<unknown>",
			acptr->user ? MaskedHost(acptr) : "<unknown>", parv[0], mypath);
		if (HasMode(sptr, UMODE_OPER)) {
			ircdlog(LOG_KILL, "KILL for %s!%s@%s. From: %s Path: %s", acptr->name,
				acptr->user ? acptr->username : "<unknown>",
				acptr->user ? MaskedHost(acptr) : "<unknown>", parv[0], mypath);
		}
		if (!MyConnect(acptr) || !MyConnect(sptr) || !HasMode(sptr, UMODE_OPER)) {
			sendto_serv_butone(cptr, ":%s KILL %s :%s", parv[0], acptr->name, mypath);
			if (chasing && IsServer(cptr)) {
				sendto_one(cptr, ":%s KILL %s :%s", me.name, acptr->name, mypath);
			}
			acptr->flags |= FLAGS_KILLED;
		}
		if (MyConnect(acptr)) {
			sendto_prefix_one(acptr, sptr, ":%s KILL %s :%s", parv[0], acptr->name, mypath);
		}
		if (MyConnect(acptr) && MyConnect(sptr) && HasMode(sptr, UMODE_OPER)) {
			ircsprintf(buf, "Local kill by %s (%s)", sptr->name,
			BadPtr(parv[2]) ? parv[0] : parv[2]);
		}
		else {
			if ((killer = strchr(mypath, '(')) == NULL) {
				killer = "()";
			}
			ircsprintf(buf, "Killed (%s %s)", sptr->name, killer);
		}
		if (exit_client(cptr, acptr, sptr, buf) == FLUSH_BUFFER) {
			return FLUSH_BUFFER;
		}
	}
	return 0;
}
