/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_map.c: MAP command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_map.c,v 1.21 2002/12/15 21:26:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_map.so",
	"/MAP command",
	4,
	"$Revision: 1.21 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("MAP", m_map, MAXPARA, MSG_NONE);
	return;
}
void _module_unload() {
	delCommand("MAP", m_map);
	return;
}
#endif

static void dump_map(aClient *cptr, aClient *server, char *mask, int prompt_length, int length)
{
	static char prompt[64];
	char *p = &prompt[prompt_length];
	int count = 0, local = 0;
	aClient *acptr;

	*p = '\0';
	if (prompt_length > 60) {
		sendto_one(cptr, rpl_str(RPL_MAPMORE), me.name, cptr->name, prompt, server->name);
	}
	else {
		for (acptr = client; acptr != NULL; acptr = acptr->next) {
			if (IsPerson(acptr) && !IsULine(acptr)) {
				count++;
				if (!mycmp(acptr->user->server, server->name)) {
					local++;
				}
			}
		}
		sendto_one(cptr, rpl_str(RPL_MAP), me.name, cptr->name, prompt, length, server->name, local,
			(local * 100) / count);
		count = 0;
	}
	if (prompt_length > 0) {
		p[-1] = ' ';
		if (p[-2] == '`') {
			p[-2] = ' ';
		}
	}
	if (prompt_length > 60) {
		return;
	}
	strcpy(p, "|-");

	for (acptr = client; acptr != NULL; acptr = acptr->next) {
		if (!IsServer(acptr) || mycmp(acptr->serv->up, server->name) ||
		  (HIDEULINES == 1 && IsULine(acptr) && !HasMode(cptr, UMODE_OPER))) {
			continue;
		}
		if (match(mask, acptr->name)) {
			acptr->serv->sflags &= ~SFLAGS_MAP;
		}
		else {
			acptr->serv->sflags |= SFLAGS_MAP;
			count++;
		}
	}
	for (acptr = client; acptr != NULL; acptr = acptr->next) {
		if (!IsServer(acptr) || mycmp(acptr->serv->up, server->name) ||
		  !(acptr->serv->sflags & SFLAGS_MAP) || (HIDEULINES == 1 && IsULine(acptr) &&
		  !HasMode(cptr, UMODE_OPER))) {
			continue;
		}
		if (--count == 0) {
			*p = '`';
		}
		dump_map(cptr, acptr, mask, prompt_length + 2, length - 2);
	}
	if (prompt_length > 0) {
		p[-1] = '-';
	}
	return;
}

/*
 * m_map
 *	parv[0] = sender prefix
 *	parv[1] = server mask
 */
int m_map(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr;
	int longest = strlen(me.name);
	static time_t last_used = 0L;

	if (ALLOWMAP == 0) {
		sendto_one(sptr, err_str(ERR_FUNCDISABLED), me.name, parv[0], "MAP");
		return 0;
	}
	if (!HasMode(sptr, UMODE_OPER)) {
		if (FLATTEN_LINKS == 1) {
			sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0], "MAP");
			return 0;
		}
		if (last_used + PACETIME_INTENSE > timeofday) {
			sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, parv[0]);
			return 0;
		}
		else {
			last_used = timeofday;
		}
		if (SPYNOTICES == 1 && IsPerson(sptr)) {
			sendto_realops_lev(SPY_LEV, "MAP requested by %s (%s@%s)", sptr->name,
				sptr->username, MaskedHost(sptr));
		}
	}
	for (acptr = client; acptr != NULL; acptr = acptr->next) {
		if (IsServer(acptr) && (strlen(acptr->name) + acptr->hopcount * 2) > longest) {
			longest = strlen(acptr->name) + acptr->hopcount * 2;
		}
	}
	if (longest > 60) {
		longest = 60;
	}
	longest += 2;
	dump_map(sptr, &me, (parc < 2) ? "*" : parv[1], 0, longest);
	sendto_one(sptr, rpl_str(RPL_MAPEND), me.name, parv[0]);
	return 0;
}
