/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_message.c: PRIVMSG and NOTICE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_message.c,v 1.34 2002/12/15 21:26:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

extern char *canonize(char *);
extern int is_silenced(aClient *, aClient *);
#ifdef FLUD
extern int check_for_flud(aClient *, aClient *, aChannel *chptr, int);
#endif

#define CTCP_NONE	0
#define CTCP_YES	1
#define CTCP_DCC	2
#define CTCP_DCCSEND	3

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_message.so",
	"/PRIVMSG and /NOTICE commands",
	4,
	"$Revision: 1.34 $",
	1, NULL, NULL
};
void _module_load() {
	addCommand("PRIVMSG", m_private, MAXPARA, MSG_RESETIDLE);
	addCommand("NOTICE", m_notice, MAXPARA, MSG_NONE);
	RequiredModules.m_message = 1;
	return;
}
void _module_unload(void) {
	delCommand("PRIVMSG", m_private);
	delCommand("NOTICE", m_notice);
	RequiredModules.m_message = 0;
	return;
}
#endif

static inline int check_for_client_flood(int notice, aClient *sptr, aClient *acptr)
{
	int delta = 0;

	if (FLOODCOUNT < 1 || sptr->user == NULL || acptr->user == 0 || IsULine(acptr) || IsServer(acptr)) {
		return 0;
	}
	if (acptr->user->first_message_time + 1 < timeofday) {
		delta = timeofday - acptr->user->first_message_time;
		acptr->user->number_of_messages -= delta;
		acptr->user->first_message_time = timeofday;

		if (acptr->user->number_of_messages <= 0) {
			acptr->user->number_of_messages = 0;
			acptr->user->flood_notice = 0;
		}
	}
	if (acptr->user->number_of_messages >= FLOODCOUNT || acptr->user->flood_notice == 1) {
		if (acptr->user->flood_notice == 0) {
			sendto_realops_lev(FLOOD_LEV, "Possible flooder: %s (%s@%s) target: %s (%s@%s)",
				sptr->name, sptr->username, MaskedHost(sptr),
				acptr->name, acptr->username, MaskedHost(acptr));
			acptr->user->flood_notice = 1;
			acptr->user->number_of_messages += 2;
		}
		if (MyClient(sptr) && notice == 0) {
			sendto_one(sptr, ":%s NOTICE %s :Messages to %s throttled due to flooding",
				me.name, sptr->name, acptr->name);
		}
		return 1;
	}
	else {
		acptr->user->number_of_messages++;
	}
	return 0;
}

static inline int check_for_channel_flood(int notice, aClient *sptr, aChannel *chptr)
{
	int delta = 0;

	if (FLOODCOUNT < 1 || sptr->user == NULL || chptr == 0 || IsULine(sptr) || IsServer(sptr)) {
		return 0;
	}
	if (chptr->first_message_time + 1 < timeofday) {
		delta = timeofday - chptr->first_message_time;
		chptr->number_of_messages -= delta;
		chptr->first_message_time = timeofday;

		if (chptr->number_of_messages <= 0) {
			chptr->number_of_messages = 0;
			chptr->flood_notice = 0;
		}
	}
	if (chptr->number_of_messages >= FLOODCOUNT || chptr->flood_notice == 1) {
		if (chptr->flood_notice == 0) {
			sendto_realops_lev(FLOOD_LEV, "Possible flooder: %s (%s@%s) target: %s",
				sptr->name, sptr->username, MaskedHost(sptr), chptr->chname);
			chptr->flood_notice = 1;
			chptr->number_of_messages += 2;
		}
		if (MyClient(sptr) && notice == 0) {
			sendto_one(sptr, ":%s NOTICE %s :Messages to %s throttled due to flooding",
				me.name, sptr->name, chptr->chname);
		}
		return 1;
	}
	else {
		chptr->number_of_messages++;
	}
	return 0;
}

static inline int check_for_ctcp(char *str, char **dccptr)
{
	if (*str == 1) {
		if (!myncmp(&str[1], "DCC", 3)) {
			if (dccptr != NULL) {
				*dccptr = &str[1];
			}
			if (!myncmp(&str[4], " SEND", 5)) {
				return CTCP_DCCSEND;
			}
			return CTCP_DCC;
		}
		if (myncmp(&str[1], "ACTION", 6)) {
			return CTCP_YES;
		}
	}
	return CTCP_NONE;
}

static inline int check_dcc_send(aClient *from, aClient *to, char *filename)
{
	ConfigItem_ban *ban = NULL;
	Link *lp;

	if (from == to || !IsPerson(from) || HasMode(from, UMODE_OPER) || !MyClient(to) || BadPtr(filename)) {
		return 0;
	}
	if ((ban = find_ban(BAN_FILE, filename)) == NULL) {
		return 0;
	}
	for (lp = to->user->dccallow; lp != NULL; lp = lp->next) {
		if (lp->flags == DCC_LINK_ME && lp->value.cptr == from) {
			return 0;
		}
	}
	sendto_one(from, ":%s NOTICE %s :*** The DCC SEND of file %s is not permitted on this server (%s), and "
		"user %s is not currently accepting files from you. Therefore, the transfer has been blocked.",
		me.name, from->name, filename, ban->reason, to->name);
	sendto_one(to, ":%s NOTICE %s :*** %s attempted to transfer a file called %s to you via DCC SEND, which"
		"is not permitted on this server. (%s)", me.name, to->name, from->name, filename, ban->reason);
	if (!(to->user->uflags & UFLAGS_DCCNOTICE)) {
		to->user->uflags |= UFLAGS_DCCNOTICE;
		sendto_one(to, ":%s NOTICE %s :*** The majority of files sent of this type contain malicious "
			"virri and trojans. In order to protect clients on this server, transfer of these "
			"is denied.", me.name, to->name);
		sendto_one(to, ":%s NOTICE %s :*** If you trust %s, and want to be able to receive banned "
			"files from him/her, you can add them to your DCC allow list. Type /DCCALLOW HELP for "
			"more information.", me.name, to->name, from->name);
		sendto_one(to, ":%s NOTICE %s :*** You will not see this message again.", me.name, to->name);
	}
	sendto_realops_lev(DCCSEND_LEV, "%s (%s@%s) attempted to send banned file %s to %s (%s@%s).", from->name,
		from->username, MaskedHost(from), filename, to->name, to->username, MaskedHost(to));
	return 1;
}

static inline int check_message_accept(aClient *from, aClient *to)
{
	Link *lp;

	if (from == to || !IsPerson(from) || HasMode(from, UMODE_OPER) || !MyClient(to) ||
	  !HasMode(to, UMODE_PRIVACY)) {
		return 0;
	}
	for (lp = to->user->accept; lp != NULL; lp = lp->next) {
		if (lp->flags == ACCEPT_LINK_ME && lp->value.cptr == from) {
			return 0;
		}
	}
	sendto_one(from, ":%s NOTICE %s :*** %s is currently in privacy mode (+p), and is not accepting "
		"private messages from you.", me.name, from->name, to->name);
	if (timeofday - to->user->last_accept_notice >= ACCEPTNOTICETIME) {
		to->user->last_accept_notice = timeofday;
		sendto_one(from, ":%s NOTICE %s :*** They have been informed of your message attempt.",
			me.name, from->name);
		sendto_one(to, ":%s NOTICE %s :*** %s is attempting to send you a private message. However, "
			"you are currently in privacy mode (+p), and so the message was blocked.", me.name,
			to->name, from->name);
		sendto_one(to, ":%s NOTICE %s :*** If you want to receive private messages from %s, you can "
			"add them to your message accept list. Type /ACCEPT HELP for more information.",
			me.name, to->name, from->name);
	}
	return 1;
}

/*
 * parse_message
 *	parv[0] = sender prefix
 *	parv[1] = targets
 *	parv[2] = message
 */
static inline int parse_message(aClient *cptr, aClient *sptr, int parc, char *parv[], int notice)
{
	aClient *acptr = NULL;
	aChannel *chptr;
	char *command = notice ? "NOTICE" : "PRIVMSG";
	char *p = NULL, *s = NULL, *s2 = NULL;
	char *nick, *server, *dccmsg;
	int i, ret, ischan = 0, flag = 0;

	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NORECIPIENT), me.name, parv[0], command);
		return 0;
	}
	if (parc < 3 || *parv[2] == '\0') {
		sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0], command);
		return 0;
	}

	if (MyConnect(sptr)) {
#ifndef ANTI_SPAMBOT_WARN_ONLY
		if (ANTI_SPAMBOT_CHECKS == 1 && sptr->join_leave_count >= MAX_JOIN_LEAVE_COUNT) {
			return 0;
		}
#endif
		parv[1] = canonize(parv[1]);
	}
	for (nick = strtoken(&p, parv[1], ","), i = 0; nick != NULL && i < 20; nick = strtoken(&p, NULL, ",")) {
		if (i++ > MAXTARGETS && !IsULine(sptr)) {
			if ((NOOPERFLOOD == 1 && !HasMode(sptr, UMODE_OPER)) || NOOPERFLOOD == 0) {
				sptr->since += 4;
			}
		}
		if ((ischan = IsChannelName(nick)) != 0 && (chptr = find_channel(nick, NULL)) != NULL) {
			if (notice == 0) {
				switch (check_for_ctcp(parv[2], NULL)) {
					case CTCP_NONE:
						break;
					case CTCP_DCCSEND:
					case CTCP_DCC:
						sendto_one(sptr, ":%s NOTICE %s :You may not send a DCC "
							"command to a channel (%s)", me.name, parv[0],
							nick);
						continue;
					default:
#ifdef FLUD
						if (check_for_flud(sptr, NULL, chptr, 1)) {
							return 0;
						}
#endif
						if (chptr->mode.mode & MODE_NOCTCP) {
							sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN),
								me.name, parv[0], nick);
							continue;
						}
						break;
				}
			}
			if ((ret = IsULine(sptr) ? 0 : can_send(sptr, chptr, parv[2])) > 0 && MyClient(sptr)) {
				if (ret == ERR_NOCOLOURSONCHAN) {
					sendto_one(sptr, err_str(ERR_NOCOLOURSONCHAN), me.name, parv[0], nick,
						parv[2]);
					continue;
				}
			}
			if (ret == 0) {
				char *text = parv[2];
				if (MyClient(sptr)) {
					if (check_for_channel_flood(notice, sptr, chptr)) {
						continue;
					}
					if (chptr->mode.mode & MODE_NOCOLOUR && msg_has_colour(text) == 1) {
						if (!notice) {
							sendto_one(sptr, err_str(ERR_NOCOLOURSONCHAN), me.name,
								parv[0], nick, parv[2]);
						}
						continue;
					}
					if (chptr->mode.mode & MODE_STRIPCOLOUR) {
						text = (char *)strip_colour(text);
					}
				}
				sendto_channel_butone(cptr, sptr, chptr, ":%s %s %s :%s", parv[0],
					command, nick, text);
			}
			else if (notice == 0) {
				sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0], nick);
			}
			continue;
		}
		if ((acptr = find_person(nick, NULL)) != NULL && ischan == 0) {
			if (notice == 0 &&
#ifdef FLUD
			  MyFludConnect(acptr)) {
#else
			  MyConnect(acptr)) {
#endif
				switch (check_for_ctcp(parv[2], &dccmsg)) {
					case CTCP_NONE:
						break;
					case CTCP_DCCSEND:
#ifdef FLUD
						if (check_for_flud(sptr, acptr, NULL, 1)) {
							return 0;
						}
#endif
						if (check_dcc_send(sptr, acptr, dccmsg)) {
							continue;
						}
						break;
					default:
#ifdef FLUD
						if (check_for_flud(sptr, acptr, NULL, 1)) {
							return 0;
						}
#endif
						break;
				}
			}
			if (!HasMode(sptr, UMODE_OPER) && !IsULine(sptr) && !IsServer(sptr)) {
				if (check_message_accept(sptr, acptr)) {
					continue;
				}
			}
			if (MyClient(sptr) && check_for_client_flood(notice, sptr, acptr)) {
				continue;
			}
			if (!is_silenced(sptr, acptr)) {
				if (!notice && MyClient(acptr) && acptr->user && acptr->user->away) {
					sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0], acptr->name,
						acptr->user->away);
				}
				sendto_prefix_one(acptr, sptr, ":%s %s %s :%s", parv[0], command, nick,
					parv[2]);
			}
			continue;
		}
		if ((s = strchr(nick, '#')) != NULL) {
			if ((chptr = find_channel(s, NULL)) == NULL) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], s);
				continue;
			}
			if ((ret = IsULine(sptr) ? 0 : can_send(sptr, chptr, parv[2])) > 0 && MyClient(sptr)) {
				if (ret == ERR_NOCOLOURSONCHAN) {
					sendto_one(sptr, err_str(ERR_NOCOLOURSONCHAN), me.name, parv[0], s, parv[2]);
					continue;
				}
			}
			if (ret > 0 && !IsULine(sptr)) {
				sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0], s);
				continue;
			}
			if (MyClient(sptr) && check_for_channel_flood(notice, sptr, chptr)) {
				continue;
			}
			if (s == nick) {
				sendto_channel_butone(cptr, sptr, chptr, ":%s %s %s :%s", parv[0], command,
					nick, parv[2]);
			}
			else {
				flag = 0;
				for (s2 = nick; s2 != s; s2++) {
					switch (*s2) {
						case '*':
							flag |= CHFL_CHANADMIN;
							break;
						case '@':
							flag |= CHFL_CHANOP;
							break;
						case '%':
							flag |= CHFL_HALFOP;
							break;
						case '+':
							flag |= CHFL_VOICE;
							break;
						default:
							break;
					}
				}
				sendto_channelflag_butone(flag, cptr, sptr, chptr, ":%s %s %s :%s", parv[0],
					command, nick, parv[2]);
			}
			continue;
		}
		if (HasMode(sptr, UMODE_OPER)) {
			if (*nick == '$' || *nick == '#') {
				if ((s = strrchr(nick, '.')) == NULL) {
					sendto_one(sptr, err_str(ERR_NOTOPLEVEL), me.name, parv[0], nick);
					continue;
				}
				while (*++s) {
					if (*s == '.' || *s == '*' || *s == '?') {
						break;
					}
				}
				if (*s == '*' || *s == '?') {
					sendto_one(sptr, err_str(ERR_WILDTOPLEVEL), me.name, parv[0], nick);
					continue;
				}
				sendto_match_butone(IsServer(cptr) ? cptr : NULL, sptr, nick + 1,
					*nick == '#' ? MATCH_HOST : MATCH_SERVER, ":%s %s %s :%s", parv[0],
					command, nick, parv[2]);
				continue;
			}
		}
		if ((server = strchr(nick, '@')) != NULL && ischan == 0) {
			if (*(server + 1) != '\0') {
				if ((acptr = find_server(server + 1, NULL)) != NULL) {
					if (!IsMe(acptr)) {
						sendto_one(acptr, ":%s %s %s :%s", parv[0], command, nick,
							parv[2]);
						continue;
					}
					if ((acptr = find_person(nick, NULL)) != NULL) {
						sendto_prefix_one(acptr, sptr, ":%s %s %s :%s", parv[0],
							command, nick, parv[2]);
					}
				}
			}
			if (acptr != NULL) {
				continue;
			}
		}
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
	}
	if (i > MAXTARGETS && sptr->user != NULL) {
		sendto_realops_lev(SPY_LEV, "%s (%s@%s) tried to message %d users", sptr->name, sptr->username, 
			MaskedHost(sptr), i);
	}
	return 0;
}

/*
 * m_private
 *	parv[0] = sender prefix
 *	parv[1] = targets
 *	parv[2] = message
 */
int m_private(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	return parse_message(cptr, sptr, parc, parv, 0);
}

/*
 * m_notice
 *	parv[0] = sender prefix
 *	parv[1] = targets
 *	parv[2] = message
 */
int m_notice(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	return parse_message(cptr, sptr, parc, parv, 1);
}
