/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_nick.c: NICK command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_nick.c,v 1.29 2002/12/15 21:26:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

extern int do_user(char *, aClient *, aClient *, char *, char *, char *, unsigned long, unsigned int, char *);
extern int register_user(aClient *, aClient *, char *, char *);
/*extern int remove_accept_references(aClient *, int);*/
extern int user_modes[];

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_nick.so",
	"/NICK command",
	4,
	"$Revision: 1.29 $",
	1, NULL, NULL
};
void _module_load() {
	addCommand("NICK", m_nick, MAXPARA, MSG_UNREGUSE);
	RequiredModules.m_nick = 1;
	return;
}
void _module_unload(void) {
	delCommand("NICK", m_nick);
	RequiredModules.m_nick = 0;
	return;
}
#endif

static int do_nick_name(char *nick)
{
	char *ch;

	if (*nick == '-' || isdigit(*nick)) {
		return 0;
	}

	for (ch = nick; *ch && (ch - nick) < NICKLEN; ch++) {
		if (!isvalid(*ch) || isspace(*ch)) {
			break;
		}
	}
	*ch = '\0';
	return (ch - nick);
}

/*
 * m_nick
 *	parv[0] = sender prefix
 *	parv[1] = nickname
 *	parv[2] = hopcount when new user; TS when nick change
 *	parv[3] = TS
 *      --- new user only below ---
 *	parv[4] = usermodes
 *	parv[5] = username
 *	parv[6] = hostname
 *	parv[7] = server
 *	parv[8] = services id
 *	if link is NICKIP:
 *		parv[9] = client ip
 *		parv[10] = realname
 *      else
 *		parv[9] = realname
 */
int m_nick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr, *uplink;
	char nick[NICKLEN + 2];
	chanMember *cm;
	ts_val newts = 0;
	int sameuser = 0, samenick = 0;

	if (parc < 2) {
		sendto_one(sptr, err_str(ERR_NONICKNAMEGIVEN), me.name, parv[0]);
		return 0;
	}
	if (!IsServer(sptr) && IsServer(cptr) && parc > 2) {
		newts = atol(parv[2]);
	}
	else if (IsServer(sptr) && parc > 3) {
		newts = atol(parv[3]);
	}
	else {
		parc = 2;
	}
	if ((IsServer(sptr) || parc > 4) && ((HasCapab(sptr, CAPAB_NICKIP) && parc < 11) ||
	  (!HasCapab(sptr, CAPAB_NICKIP) && parc < 10))) {
		sendto_realops_lev(DEBUG_LEV, "IGNORING BAD NICK: %s[%s@%s] on %s (from %s)",
			parv[1], parc >= 6 ? parv[5] : "-", parc >= 7 ? parv[6] : "-",
			parc >= 8 ? parv[7] : "-", parv[0]);
		return 0;
	}

	strncpyzt(nick, parv[1], NICKLEN + 1);
	if (!do_nick_name(nick) || (IsServer(cptr) && irccmp(nick, parv[1]))) {
		sendto_one(sptr, err_str(ERR_ERRONEUSNICKNAME), me.name, parv[0], parv[1]);
		if (IsServer(cptr)) {
			ircstp->is_kill++;
			sendto_realops_lev(DEBUG_LEV, "Bad nick: %s From: %s %s", parv[1], parv[0],
				get_client_name(cptr, FALSE));
			sendto_one(cptr, ":%s KILL %s :Bad nick", me.name, parv[1]);
			if (sptr != cptr) {
				sendto_serv_butone(cptr, ":%s KILL %s :Bad nick", me.name, parv[0]);
				sptr->flags |= FLAGS_KILLED;
				return exit_client(cptr, sptr, &me, "Bad nick");
			}

		}
		return 0;
	}
	do {
		if ((acptr = find_server(nick, NULL)) != NULL) {
			if (MyConnect(sptr)) {
				sendto_one(sptr, err_str(ERR_NICKNAMEINUSE), me.name,
					BadPtr(parv[1]) ? "*" : parv[0], nick);
				return 0;
			}
			sendto_realops_lev(SKILL_LEV, "Nick collision on %s", sptr->name);
			ircstp->is_kill++;
			sendto_one(cptr, ":%s KILL %s :Nick vollision", me.name, sptr->name);
			sptr->flags |= FLAGS_KILLED;
			return exit_client(cptr, sptr, &me, "Nick/Server collision");
		}
		if ((acptr = find_client(nick, NULL)) == NULL) {
			break;
		}
		if (acptr == sptr) {
			if (irccmp(acptr->name, nick)) {
				break;
			}
			else {
				return 0;
			}
		}
		if (IsUnknown(acptr)) {
			if (MyConnect(acptr)) {
				exit_client(NULL, acptr, &me, "Overridden");
				break;
			}
			else if (acptr->user == NULL) {
				sendto_realops_lev(SKILL_LEV, "Nick collision on %s", parv[1]);
				sendto_serv_butone(NULL, ":%s KILL %s :Nick collision", me.name, acptr->name);
				acptr->flags |= FLAGS_KILLED;
				return exit_client(cptr, acptr, &me, "Got TS NICK before NON-TS USER");
			}
		}
		if (!IsServer(cptr)) {
			sendto_one(sptr, err_str(ERR_NICKNAMEINUSE), me.name,
				BadPtr(parv[0]) ? "*" : parv[0], nick);
			return 0;
		}
		if (IsServer(sptr)) {
			if (newts == 0 || acptr->tsinfo == 0 || newts == acptr->tsinfo) {
				sendto_realops_lev(SKILL_LEV, "Nick collision on %s", parv[1]);
				ircstp->is_kill++;
				sendto_one(acptr, err_str(ERR_NICKCOLLISION), me.name, acptr->name,
					acptr->name);
				sendto_serv_butone(NULL, ":%s KILL %s :Nick collision", me.name, acptr->name);
				acptr->flags |= FLAGS_KILLED;
				return exit_client(cptr, acptr, &me, "Nick collision");
			}
			else {
				sameuser = ((acptr->user != NULL) && !mycmp(acptr->username, parv[5]) &&
					!mycmp(acptr->host, parv[6]));
				if ((sameuser && newts < acptr->tsinfo) || (!sameuser && newts > acptr->tsinfo)) {
					return 0;
				}
				else {
					sendto_realops_lev(SKILL_LEV, "Nick collision on %s", parv[1]);
					ircstp->is_kill++;
					sendto_one(acptr, err_str(ERR_NICKCOLLISION), me.name, acptr->name,
						acptr->name);
					sendto_serv_butone(sptr, ":%s KILL %s :Nick collision", me.name, acptr->name);
					acptr->flags |= FLAGS_KILLED;
					exit_client(cptr, acptr, &me, "Nick collision");
					break;
				}
			}
		}
		if (newts == 0 || acptr->tsinfo == 0 || newts == acptr->tsinfo || sptr->user == NULL) {
			sendto_realops_lev(SKILL_LEV, "Nick change collision: %s", parv[1]);
			ircstp->is_kill++;
			sendto_one(acptr, err_str(ERR_NICKCOLLISION), me.name, acptr->name, acptr->name);
			sendto_serv_butone(NULL, ":%s KILL %s :Nick collision", me.name, sptr->name);
			ircstp->is_kill++;
			sendto_one(sptr, err_str(ERR_NICKCOLLISION), me.name, sptr->name, sptr->name);
			sendto_serv_butone(NULL, ":%s KILL %s :Nick collision", me.name, acptr->name);
			acptr->flags |= FLAGS_KILLED;
			exit_client(NULL, acptr, &me, "Nick collision (new)");
			sptr->flags |= FLAGS_KILLED;
			return exit_client(cptr, sptr, &me, "Nick collision (old)");
		}
		else {
			sameuser = (!mycmp(acptr->username, sptr->username) && !mycmp(acptr->host, sptr->host));
			if ((sameuser && newts < acptr->tsinfo) || (!sameuser && newts > acptr->tsinfo)) {
				if (sameuser) {
					sendto_realops_lev(SKILL_LEV, "Nick change collision from %s to %s",
						sptr->name, acptr->name);
				}
				ircstp->is_kill++;
				sendto_serv_butone(cptr, ":%s KILL %s :Nick collision", me.name, sptr->name);
				sptr->flags |= FLAGS_KILLED;
				return exit_client(cptr, sptr, &me,
					sameuser ? "Nick collision(old)" : "Nick collision (new)");
			}
			else {
				sendto_realops_lev(SKILL_LEV, "Nick collision on %s", acptr->name);
				ircstp->is_kill++;
				sendto_one(acptr, err_str(ERR_NICKCOLLISION), me.name, acptr->name,
					acptr->name);
				sendto_serv_butone(sptr, ":%s KILL %s :Nick collision", me.name, acptr->name);
				acptr->flags |= FLAGS_KILLED;
				exit_client(cptr, acptr, &me, "Nick collision");
			}
		}
	} while (0);

	if (IsServer(sptr)) {
		if ((uplink = find_server(parv[7], NULL)) == NULL) {
			sendto_realops("Remote nick %s on UNKNOWN server %s", nick, parv[7]);
			return 0;
		}
		sptr = make_client(cptr, uplink);

		if (IsULine(uplink)) {
			sptr->flags |= FLAGS_ULINE;
		}

		add_client_to_list(sptr);
		if (parc > 2) {
			sptr->hopcount = atoi(parv[2]);
		}
		if (newts > 0) {
			sptr->tsinfo = newts;
		}
		else {
			newts = sptr->tsinfo = (ts_val)timeofday;
			ts_warn("Remote nick %s introduced without a TS", nick);
		}

		strcpy(sptr->name, nick);
		add_to_client_hash_table(nick, sptr);

		if (parc >= 10) {
			int *s, flag;
			char *m;

			if (parc == 11 && !HasCapab(cptr, CAPAB_NICKIP)) {
				sendto_realops_lev(DEBUG_LEV, "Got extra parameters from non-NICKIP server %s",
					cptr->name);
			}
			for (m = &parv[4][1]; *m; m++) {
				for (s = user_modes; (flag = *s); s += 2) {
					if (*m == (char)(*(s + 1))) {
						if (flag == UMODE_OPER) {
							Count.oper++;
						}
						sptr->umode |= flag & SEND_UMODES;
						break;
					}
				}
			}
			if (parc == 10) {
				return do_user(nick, cptr, sptr, parv[5], parv[6], parv[7],
					strtoul(parv[8], NULL, 0), 0, parv[9]);
			}
			else if (parc == 11) {
				return do_user(nick, cptr, sptr, parv[5], parv[6], parv[7],
					strtoul(parv[8], NULL, 0), strtoul(parv[9], NULL, 0),
					parv[10]);
			}
		}
	}
	else if (*sptr->name != '\0')
	{
		int remove_regnick = 0;
		ConfigItem_ban *ban = NULL;

		if (mycmp(parv[0], nick)) {
			remove_regnick = 1;
			sptr->tsinfo = newts ? newts : (ts_val)timeofday;
		}
		if ((ban = find_ban(BAN_NICKNAME, nick)) != NULL) {
			if (!MyConnect(sptr)) {
				sendto_realops("Q:Lined nick %s from %s on %s", nick,
					*sptr->name && IsServer(sptr) ? sptr->name : "<unregistered>",
					!sptr->user ? IsServer(sptr) ? parv[6] : me.name : sptr->user->server);
			}
			if (MyConnect(sptr) && !IsServer(cptr) && !HasMode(cptr, UMODE_OPER) && !IsULine(cptr)) {
				sendto_one(sptr, err_str(ERR_ERRONEUSNICKNAME), me.name,
					BadPtr(parv[0]) ? "*" : parv[0], nick, ban->reason, ban->mask);
				sendto_realops("Forbidding Q:Lined nickname %s from %s.", nick,
					get_client_name(cptr, FALSE));
				return 0;
			}
		}
		if (MyConnect(sptr)) {
			if (IsRegisteredUser(sptr)) {
				for (cm = sptr->user->channel; cm != NULL; cm = cm->nextchan) {
					if (can_send(sptr, cm->chptr, NULL)) {
						sendto_one(sptr, err_str(ERR_BANNICKCHANGE), me.name,
							sptr->name, cm->chptr->chname);
						return 0;
					}
					if (is_banned(sptr, cm->chptr)) {
						sendto_one(sptr, err_str(ERR_BANONCHAN), me.name,
							sptr->name, nick, cm->chptr->chname);
						return 0;
					}
					if (cm->chptr->mode.mode & MODE_NONICKCH && !IsULine(cptr) &&
					  !is_chan_admin(cptr, cm->chptr)) {
						sendto_one(sptr, ":%s NOTICE %s :*** Cannot change nickname: channel "
							"%s is set +N", me.name, sptr->name, cm->chptr->chname);
						return 0;
					}
				}
				if (ANTINICKFLOOD == 1) {
					if (sptr->user->last_nick_change + ANTINICKFLOOD_TIME < timeofday) {
						sptr->user->nick_changes = 0;
					}
					sptr->user->last_nick_change = timeofday;
					sptr->user->nick_changes++;

					if (!HasMode(sptr, UMODE_OPER) &&
					  sptr->user->nick_changes > ANTINICKFLOOD_CHANGES) {
						sendto_one(sptr, ":%s NOTICE %s :*** Too many nickname changes. "
							"Please wait %d seconds before trying again.", me.name,
							sptr->name, ANTINICKFLOOD_TIME);
						return 0;
					}
				}

				sendto_common_channels(sptr, ":%s NICK :%s", parv[0], nick);
				if (sptr->user != NULL) {
					add_history(sptr, 1);
					sendto_serv_butone(cptr, ":%s NICK %s :%ld", parv[0], nick, sptr->tsinfo);
				}
			}
		}
		else {
			sendto_common_channels(sptr, ":%s NICK :%s", parv[0], nick);
			if (sptr->user != NULL) {
				add_history(sptr, 1);
				sendto_serv_butone(cptr, ":%s NICK %s :%ld", parv[0], nick, sptr->tsinfo);
			}
		}
		if (remove_regnick == 1) {
			DelMode(sptr, UMODE_REGNICK);
		}
	}
	else {
		if (MyConnect(sptr)) {
			ConfigItem_ban *ban = NULL;
			if ((ban = find_ban(BAN_NICKNAME, nick)) != NULL) {
				if (MyConnect(sptr) && !IsServer(sptr) && !HasMode(sptr, UMODE_OPER) && !IsULine(cptr)) {
					sendto_one(sptr, err_str(ERR_ERRONEUSNICKNAME), me.name,
						BadPtr(parv[0]) ? "*" : parv[0], nick, ban->reason, ban->mask);
					sendto_realops("Forbidding Q:Lined nick %s from <unregistered>%s.",
						nick, get_client_name(cptr, FALSE));
					return 0;
				}
			}
		}

		strcpy(sptr->name, nick);
		sptr->tsinfo = timeofday;
		if (sptr->user != NULL) {
			if (register_user(cptr, sptr, nick, sptr->username) == FLUSH_BUFFER) {
				return FLUSH_BUFFER;
			}
		}
	}
	if (*sptr->name != '\0') {
		del_from_client_hash_table(sptr->name, sptr);
		samenick = mycmp(sptr->name, nick) ? 0 : 1;
		if (IsPerson(sptr) && samenick == 0) {
			hash_check_watch(sptr, RPL_LOGOFF);
		}
	}

	strcpy(sptr->name, nick);
	add_to_client_hash_table(nick, sptr);

	if (IsPerson(sptr) && samenick == 0) {
		hash_check_watch(sptr, RPL_LOGON);
	}
	return 0;
}
