/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_server.c: SERVER protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_server.c,v 1.42 2002/12/15 21:26:18 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "patchlevel.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

extern int do_server_estab(aClient *);

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_server.so",
	"SERVER protocol",
	4,
	"$Revision: 1.42 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("SERVER", m_server, MAXPARA, MSG_UNREGUSE);
	return;
}
void _module_unload(void) {
	delCommand("SERVER", m_server);
	return;
}
#endif

/*
 * m_server
 *	parv[0] = sender prefix
 *	parv[1] = server name
 *	parv[2] = hopcount
 *	parv[3] = server information
 */
int m_server(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *servername, *s, *d, *inpath = get_client_name(cptr, HIDEME);
	char clean_servername[(HOSTLEN * 2) + 1], info[REALLEN + 61];
	ConfigItem_link *link = NULL;
	ConfigItem_ban *ban = NULL;
	aClient *acptr, *bcptr;
	int hop = 0, bogus_server = 0, dot = 0, i = 0, n;
#ifndef HAVE_HUB_ON
	int h;
#endif

	if (IsPerson(cptr)) {
		return 0;
	}
	if (cptr->listener != NULL && cptr->listener->bind != NULL) {
		if (cptr->listener->bind->options & BIND_CLIENTONLY) {
			sendto_realops_lev(REJ_LEV, "Blocking unknown connection on client-only port %d",
				cptr->listener->bind->port);
			return exit_client(cptr, sptr, &me, "This port is for clients only");
		}
	}
	if (parc < 3) {
		sendto_one(cptr, "ERROR :Not enough SERVER parameters");
		return exit_client(cptr, sptr, &me, "Not enough SERVER parameters");
	}

	servername = parv[1];
	if (strlen(servername) > HOSTLEN) {
		servername[HOSTLEN] = '\0';
	}

	hop = 0;
	*info = '\0';
	if (parc > 3 && (i = atoi(parv[2]))) {
		hop = i;
		strncpyzt(info, parv[3], REALLEN);
	}
	else if (parc > 2) {
		strncpyzt(info, parv[2], REALLEN);
		if (parc > 3 && (i = strlen(info)) < REALLEN - 2) {
			strcat(info, " ");
			strncat(info, parv[3], REALLEN - i - 2);
			info[REALLEN] = '\0';
		}
	}

	s = servername;
	d = clean_servername;
	n = (HOSTLEN * 2) - 2;

	while (*s != '\0' && n > 0) {
		if ((unsigned char)*s < (unsigned char)' ') {
			bogus_server = 1;
			*d++ = '^';
			*d++ = (char)((unsigned char)*s + 0x40);
			n -= 2;
		}
		else if ((unsigned char)*s > (unsigned char)'~') {
			bogus_server = 1;
			*d++ = '.';
			n--;
		}
		else {
			if (*s == '.') {
				dot = 1;
			}
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	if (dot == 0 || bogus_server == 1) {
		sendto_one(cptr, "ERROR :Bogus server name (%s)", clean_servername);
		ircdlog(LOG_SERVER, "bogus server name: %s", clean_servername);
		return exit_client(cptr, sptr, &me, "Bogus server name");
	}
	if (BadPtr(cptr->passwd)) {
		sendto_one(cptr, "ERROR :No password");
		ircdlog(LOG_SERVER, "no password from %s", get_client_name(cptr, FALSE));
		return exit_client(cptr, sptr, &me, "No password");
	}
	if ((acptr = find_name(servername, NULL)) != NULL) {
		bcptr = cptr->firsttime > acptr->from->firsttime ? cptr : acptr->from;
		sendto_one(bcptr, "ERROR :Server %s already exists", servername);
		if (bcptr == cptr) {
			sendto_gnotice("Link %s cancelled, server %s already exists",
				inpath, servername);
			sendto_serv_butone(bcptr, ":%s GNOTICE :Link %s cancelled, server "
				"%s already exists", me.name, get_client_name(bcptr, HIDEME),
				servername);
			ircdlog(LOG_SERVER, "link %s cancelled: server %s alread exists",
				get_client_name(cptr, FALSE), get_client_name(bcptr, FALSE));
			return exit_client(bcptr, bcptr, &me, "Server exists");
		}
		sendto_gnotice("Link %s cancelled, server %s reintroduced by %s",
			inpath, servername, get_client_name(bcptr, HIDEME));
		sendto_serv_butone(bcptr, ":%s GNOTICE :Link %s cancelled, server %s "
			"reintroduced by %s", me.name, inpath, servername, get_client_name(bcptr, HIDEME));
		ircdlog(LOG_SERVER, "link %s cancelled: server %s reintroduced by %s",
			get_client_name(cptr, FALSE), servername, get_client_name(bcptr, FALSE));
		exit_client(bcptr, bcptr, &me, "Server exists");
	}
	if ((acptr = find_client(servername, NULL)) != NULL && acptr != cptr) {
		sendto_gnotice("Link %s cancelled, server/nick collision", inpath);
		sendto_one(cptr, "ERROR :Nickname %s already exists", servername);
		ircdlog(LOG_SERVER, "link %s cancelled: server/nick collision", get_client_name(cptr, FALSE));
		return exit_client(cptr, sptr, &me, "Server/nick collision");
	}
	if (IsServer(cptr)) {
		if (parc == 1 || *info == '\0') {
			sendto_one(cptr, "ERROR :No server information specified for %s", servername);
			ircdlog(LOG_SERVER, "link %s cancelled: no server information", get_client_name(cptr, FALSE));
			return 0;
		}
		if (hop > LINKDEPTH) {
			sendto_gnotice("Link %s cancelled, too many servers (hopcount exceeds linkdepth) "
				"[%d > %d]", inpath, hop, LINKDEPTH);
			sendto_one(cptr, "ERROR :Too many servers (hopcount exceeds linkdepth) [%d > %d]",
				hop, LINKDEPTH);
			ircdlog(LOG_SERVER, "link %s cancelled: too many servers (hopcount %d > linkdepth %d)",
				get_client_name(cptr, FALSE), hop, LINKDEPTH);
			return exit_client(cptr, sptr, &me, "Too many servers");
		}
		acptr = make_client(cptr, sptr);
		make_server(acptr);
		acptr->hopcount = hop;
		strncpyzt(acptr->name, servername, sizeof(acptr->name));
		strncpyzt(acptr->info, info, REALLEN);
		acptr->serv->up = find_or_add(parv[0]);
		SetServer(acptr);
		Count.server++;

		if (IsULine(sptr) || find_uline(acptr->name)) {
			sendto_gnotice("Link %s introducing U:Lined server %s", inpath, acptr->name);
			acptr->flags |= FLAGS_ULINE;
		}

		add_client_to_list(acptr);
		add_to_client_hash_table(acptr->name, acptr);

		for (i = 0; i <= highest_fd; i++) {
			if (!(bcptr = local[i]) || !IsServer(bcptr) || bcptr == cptr || IsMe(bcptr)) {
				continue;
			}
			sendto_one(bcptr, ":%s SERVER %s %d :%s", parv[0], acptr->name,
				acptr->hopcount + 1, acptr->info);
		}
		return 0;
	}
	if (!IsUnknown(cptr) && !IsHandshake(cptr)) {
		return 0;
	}

	if ((link = find_link(servername, cptr->sockhost)) == NULL) {
		sendto_gnotice("Link %s cancelled, no link configuration", inpath);
		sendto_one(cptr, "ERROR :No link configuration");
		ircdlog(LOG_SERVER, "link %s cancelled: no link configuration", get_client_name(cptr, FALSE));
		return exit_client(cptr, sptr, &me, "No link configuration");
	}
	if ((ban = find_ban(BAN_SERVER, servername)) != NULL) {
		sendto_gnotice("Link %s cancelled, server banned (%s)", inpath, ban->reason);
		sendto_one(cptr, "ERROR :Server banned (%s)", ban->reason);
		ircdlog(LOG_SERVER, "link %s cancelled: banned (%s)", get_client_name(cptr, FALSE),
			ban->reason);
		return exit_client(cptr, sptr, &me, "Server banned");
	}
	if (link->class->clients + 1 > link->class->maxclients) {
		sendto_gnotice("Link %s cancelled, class full {%s}", inpath, link->class->name);
		sendto_one(cptr, "ERROR :Class full {%s}", link->class->name);
		ircdlog(LOG_SERVER, "link %s cancelled: class full {%s}", get_client_name(cptr, FALSE),
			link->class->name);
		return exit_client(cptr, sptr, &me, "Class full");
	}
	if (link->auth == NULL) {
		sendto_gnotice("Link %s cancelled, no link %s::passwd section", link->servername);
		sendto_one(cptr, "ERORR :No link::passwd configuration");
		ircdlog(LOG_SERVER, "link %s cancelled: no link::passwd configuration");
		ircdlog(LOG_DEFAULT, "no link::passwd configuration for server %s", link->servername);
		return exit_client(cptr, sptr, &me, "No link::passwd configuration");
	}
	if (!auth_check(link->auth, cptr->passwd)) {
		sendto_gnotice("Link %s cancelled, authentication failure", inpath);
		sendto_one(cptr, "ERROR :Authentication failure");
		ircdlog(LOG_SERVER, "link %s cancelled: authentication failure", get_client_name(cptr, FALSE));
		return exit_client(cptr, sptr, &me, "Authentication failure");
	}

#ifndef HAVE_HUB_ON
	for (h = 0; h <= highest_fd; h++) {
		if (local[h] && IsServer(local[h])) {
			ircstp->is_ref++;
			sendto_one(cptr, "ERROR :I'm a leaf not a hub");
			ircdlog(LOG_SERVER, "link %s cancelled: i'm a leaf!", get_client_name(cptr, FALSE));
			return exit_client(cptr, cptr, cptr, "I'm a leaf");
		}
	}
#endif

	strncpyzt(cptr->name, servername, sizeof(cptr->name));
	strncpyzt(cptr->info, info ? info : "IRCers United", REALLEN);
	cptr->hopcount = hop;

	if (link->options & LINK_COMPRESS) {
		AddCapab(cptr, CAPAB_DOZIP);
	}
	if (link->options & LINK_SECURE) {
		AddCapab(cptr, CAPAB_DODKEY);
	}

	if (IsUnknown(cptr)) {
		sendto_one(cptr, "PASS %s :TS", link->auth->string);
		sendto_one(cptr, "CAPAB NOQUIT BURST UNCONNECT ZIP SSJv2 NICKIP VHOST SNICK");
#ifdef HAVE_OPENSSL_ON
		if (HasCapab(cptr, CAPAB_DODKEY)) {
			sendto_one(cptr, "CAPAB DKEY");
		}
#endif
		sendto_one(cptr, "SERVER %s 1 :%s", me.name, me.info ? me.info : "IRCers United");
	}

	if (!DoesTS(cptr)) {
		sendto_gnotice("Warning! Link %s is not a TS server", inpath);
		ircdlog(LOG_SERVER, "WARNING: link %s is not a TS server", get_client_name(cptr, FALSE));
	}

	sendto_one(cptr, "SVINFO %d %d 0 %ld %s %s :%ld.%ld.%ld", TS_CURRENT, TS_MIN,
		(ts_val)timeofday, CODENAME, IRCNETWORK, MASK_KEY1, MASK_KEY2, MASK_KEY3);

	make_server(cptr);
	cptr->serv->up = me.name;
	cptr->serv->conf = link;
	cptr->class = link->class;
	cptr->class->clients++;
	cptr->pingval = cptr->class->pingtime;
	cptr->sendqlen = cptr->class->sendq;
	link->refcount++;

	strncpy(cptr->hostip, (char *)inetntoa((char *)&cptr->ip), sizeof(cptr->hostip));
	strncpy(cptr->sockhost, cptr->hostip, sizeof(cptr->sockhost));

#ifdef HAVE_OPENSSL_ON
	if (HasCapab(cptr, CAPAB_DKEY) && HasCapab(cptr, CAPAB_DODKEY)) {
		cptr->serv->sflags |= SFLAGS_RC4_NEG;
		sendto_one(cptr, "DKEY START");
		return 0;
	}
#endif
	return do_server_estab(cptr);
}
