/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_whois.c: WHOIS command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_whois.c,v 1.28 2002/12/15 21:26:20 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

static char buf[BUFSIZE];
static int bufpos = 0;

extern char *get_mode_str(aClient *);

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_whois.so",
	"/WHOIS command",
	4,
	"$Revision: 1.28 $",
	0, NULL, NULL
};
void _module_load() {
	addCommand("WHOIS", m_whois, MAXPARA, MSG_NONE);
	return;
}
void _module_unload(void) {
	delCommand("WHOIS", m_whois);
	return;
}
#endif

static inline void w_addtobuf(char *string)
{
	char *ptr;
	for (ptr = string; *ptr; ptr++) {
		buf[bufpos++] = *ptr;
	}
	buf[bufpos] = '\0';
	return;
}

/*
 * m_whois
 *	parv[0] = sender prefix
 *	parv[1] = nickname list
 */
int m_whois(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr;
	aChannel *chptr;
	chanMember *cm;
	char *nick, *tmp, *name = NULL, *p = NULL;
	int found, len, showchan = 0;
	static time_t last_used = 0L;

	if (!HasMode(sptr, UMODE_OPER)) {
		if (last_used + PACETIME_SIMPLE > timeofday) {
			sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, parv[0]);
			return 0;
		}
		else {
			last_used = timeofday;
		}
	}
	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "WHOIS");
		return 0;
	}
	if (parc > 2) {
		if (hunt_server(cptr, sptr, ":%s WHOIS %s :%s", 1, parc, parv) != HUNTED_ISME) {
			return 0;
		}
		parv[1] = parv[2];
	}
	for (tmp = parv[1]; (nick = strtoken(&p, tmp, ",")); tmp = NULL) {
		found = 0;
		buf[(bufpos = 0)] = '\0';

		if ((acptr = hash_find_client(nick, NULL)) == NULL || !IsPerson(acptr)) {
			sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
			return 0;
		}
		name = *acptr->name != '\0' ? acptr->name : "?";

		sendto_one(sptr, rpl_str(RPL_WHOISUSER), me.name, parv[0], name, acptr->username,
			MaskedHost(acptr), acptr->info);
		if (HasMode(sptr, UMODE_OPER) || sptr == acptr) {
			sendto_one(sptr, rpl_str(RPL_WHOISHOSTANDMODES), me.name, parv[0], name,
				acptr->host, get_mode_str(acptr));
		}
		if (!HasMode(acptr, UMODE_SCLIENT)) {
			len = strlen(me.name) + strlen(parv[0]) + 6 + strlen(name);
			for (cm = acptr->user->channel; cm != NULL; cm = cm->nextchan) {
				chptr = cm->chptr;
				showchan = ShowChannel(sptr, chptr) ? 1 : 0;

				if (showchan == 0 && !HasMode(sptr, UMODE_SADMIN)) {
					continue;
				}
				if (bufpos + strlen(chptr->chname) > BUFSIZE - 4 - len) {
					sendto_one(sptr, ":%s %d %s %s :%s", me.name, RPL_WHOISCHANNELS,
						parv[0], name, buf);
					buf[(bufpos = 0)] = '\0';
				}
				if (bufpos > 0) {
					buf[bufpos++] = ' ';
				}
				if (showchan == 0) {
					buf[bufpos++] = '!';
				}
				if (cm->flags & CHFL_CHANADMIN) {
					buf[bufpos++] = '*';
				}
				else if (cm->flags & CHFL_CHANOP) {
					buf[bufpos++] = '@';
				}
				else if (cm->flags & CHFL_HALFOP) {
					buf[bufpos++] = '%';
				}
				else if (cm->flags & CHFL_VOICE) {
					buf[bufpos++] = '+';
				}
				w_addtobuf(chptr->chname);
			}
			if (bufpos > 0) {
				sendto_one(sptr, rpl_str(RPL_WHOISCHANNELS), me.name, parv[0], name, buf);
				buf[(bufpos = 0)] = '\0';
			}
		}
		sendto_one(sptr, rpl_str(RPL_WHOISSERVER), me.name, parv[0], name, acptr->user->server,
			acptr->uplink != NULL ? acptr->uplink->info : "*NOT ON THIS NET*");
		if (HasMode(acptr, UMODE_REGNICK)) {
			sendto_one(sptr, rpl_str(RPL_WHOISREGNICK), me.name, parv[0], name);
		}
		if (acptr->user->away != NULL) {
			sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0], name, acptr->user->away);
		}
		if (HasMode(acptr, UMODE_OPER)) {
			w_addtobuf(":%s %d %s %s :is an Operator");
			if (HasMode(acptr, UMODE_NETADMIN)) {
				w_addtobuf(" - Network Administrator");
			}
			else if (HasMode(acptr, UMODE_ADMIN)) {
				w_addtobuf(" - Server Administrator");
			}
			else if (HasMode(acptr, UMODE_COADMIN)) {
				w_addtobuf(" - Server Co-Administrator");
			}
			sendto_one(sptr, buf, me.name, RPL_WHOISOPERATOR, parv[0], name);
		}
		if (HasMode(acptr, UMODE_SADMIN)) {
			sendto_one(sptr, rpl_str(RPL_WHOISSADMIN), me.name, parv[0], name);
		}
		if (HasMode(acptr, UMODE_SCLIENT)) {
			sendto_one(sptr, rpl_str(RPL_WHOISSERVICECLI), me.name, parv[0], name);
		}
		if (acptr->flags & FLAGS_SECURE) {
			sendto_one(sptr, rpl_str(RPL_WHOISSECURE), me.name, parv[0], name);
		}
		if (MyConnect(acptr)) {
			sendto_one(sptr, rpl_str(RPL_WHOISIDLE), me.name, parv[0], name,
				timeofday - acptr->user->last, acptr->firsttime);
		}
		continue;
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFWHOIS), me.name, parv[0], name);
	return 0;
}
