/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/motd.c: motd file handling functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: motd.c,v 1.19 2002/10/25 22:42:24 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <time.h>
#include "h.h"
#include "memory.h"

MessageFile motd;
MessageFile opermotd;
MessageFile rules;

static void destroy_message_file_data(MessageFile *ptr)
{
	MessageFileLine *l, *next = NULL;
	for (l = ptr->data; l != NULL; l = next) {
		next = l->next;
		MyFree(l->line);
		MyFree(l);
	}
	ptr->data = NULL;
	*ptr->changed_date = '\0';
	return;
}

static void init_message_file(MessageFile *ptr, char *filename, short type)
{
	DupString(ptr->filename, filename);
	ptr->data = NULL;
	ptr->type = type;
	*ptr->changed_date = '\0';
	return;
}

void load_message_file(MessageFile *ptr)
{
	MessageFileLine *l, *cur = NULL;
	char buffer[128];
	int fd = -1;
	struct stat sb;
	struct tm *localtm;

	destroy_message_file_data(ptr);
	if ((fd = open(ptr->filename, O_RDONLY)) == -1) {
		return;
	}
	if (fstat(fd, &sb) != -1 && sb.st_size > 0) {
		dgets(-1, NULL, 0);
		while (dgets(fd, buffer, 128) > 0) {
			strip(buffer);
			l = (MessageFileLine *)MyMallocX(sizeof(MessageFileLine));
			l->next = NULL;
			DupString(l->line, buffer);

			if (ptr->data != NULL) {
				if (cur != NULL) {
					cur->next = l;
				}
				cur = l;
			}
			else {
				ptr->data = l;
				cur = l;
			}
		}
		if ((localtm = localtime(&sb.st_mtime)) != NULL) {
			ircsprintf(ptr->changed_date, "%d/%d/%d %d:%02d", localtm->tm_mday, localtm->tm_mon + 1,
				localtm->tm_year + 1900, localtm->tm_hour, localtm->tm_min);
		}
	}
	close(fd);
	return;
}

void send_message_file(aClient *cptr, MessageFile *ptr)
{
	MessageFileLine *l;

	if (cptr == NULL || ptr == NULL) {
		ircdlog(LOG_ERROR, "send_message_file() encountered NULL parameter, cptr %lX ptr %lX",
			cptr, ptr);
		return;
	}
	if (ptr->data == NULL) {
		sendto_one(cptr,
			(ptr->type == MFTYPE_MOTD) ? err_str(ERR_NOMOTD) : ":%s NOTICE %s :File is missing.",
			me.name, cptr->name);
		return;
	}
	switch (ptr->type) {
		case MFTYPE_MOTD:
			sendto_one(cptr, ":%s NOTICE %s :*** The MOTD was last changed %s.", me.name,
				cptr->name, ptr->changed_date);
			sendto_one(cptr, rpl_str(RPL_MOTDSTART), me.name, cptr->name, me.name);
			for (l = ptr->data; l != NULL; l = l->next) {
				sendto_one(cptr, rpl_str(RPL_MOTD), me.name, cptr->name, l->line);
			}
			sendto_one(cptr, rpl_str(RPL_ENDOFMOTD), me.name, cptr->name);
			return;
		case MFTYPE_OPERMOTD:
			sendto_one(cptr, ":%s NOTICE %s :Operator Message of the Day, last changed %s.",
				me.name, cptr->name, ptr->changed_date);
			for (l = ptr->data; l != NULL; l = l->next) {
				sendto_one(cptr, rpl_str(RPL_MOTD), me.name, cptr->name, l->line);
			}
			sendto_one(cptr, ":%s NOTICE %s :End of OPERMOTD.", me.name, cptr->name);
			return;;
		case MFTYPE_RULES:
			sendto_one(cptr, ":%s NOTICE %s :Server Rules, last changed %s.", me.name,
				cptr->name, ptr->changed_date);
			for (l = ptr->data; l != NULL; l = l->next) {
				sendto_one(cptr, rpl_str(RPL_MOTD), me.name, cptr->name, l->line);
			}
			sendto_one(cptr, ":%s NOTICE %s :End of RULES.", me.name, cptr->name);
			return;;
	}
	return;
}

void init_MessageFiles()
{
	init_message_file(&motd, MOTD_FILE, MFTYPE_MOTD);
	load_message_file(&motd);
	init_message_file(&opermotd, OPERMOTD_FILE, MFTYPE_OPERMOTD);
	load_message_file(&opermotd);
	init_message_file(&rules, RULES_FILE, MFTYPE_RULES);
	load_message_file(&rules);
	return;
}
