/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/zlink.c: zipped links
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: zlink.c,v 1.10 2002/11/20 22:26:36 rageircd Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include "zlink.h"
#include "memory.h"

extern void ircdlog(char *, char *, ...);

static char zipOutBuf[zipOutBufSize];
static char zipInBuf[zipInBufSize];

void *zip_create_input_session()
{
	ZLinkIn *z = (ZLinkIn *)MyMallocX(sizeof(ZLinkIn));
	z->stream.zalloc = NULL;
	z->stream.zfree = NULL;
	z->stream.data_type = Z_ASCII;
	if (inflateInit(&z->stream) != Z_OK) {
		return NULL;
	}
	return (void *)z;
}

void *zip_create_output_session()
{
	ZLinkOut *z = (ZLinkOut *)MyMallocX(sizeof(ZLinkOut));
	z->stream.zalloc = NULL;
	z->stream.zfree = NULL;
	z->stream.data_type = Z_ASCII;
	if (deflateInit(&z->stream, COMPRESSION_LEVEL) != Z_OK) {
		return NULL;
	}
	return (void *)z;
}

char *zip_input(void *session, char *buffer, int *len, int *err, char **nbuf, int *nbuflen)
{
	ZLinkIn *z;
	z_stream *zin;
	int ret = 0;

	if ((z = (ZLinkIn *)session) == NULL) {
		ircdlog(LOG_ERROR, "zip_input() with NULL session pointer");
		*len = -1;
		*err = ret;
		return "????";
	}

	*nbuf = NULL;
	*err = 0;
	zin = &z->stream;
	zin->next_in = buffer;
	zin->avail_in = *len;
	zin->next_out = zipInBuf;
	zin->avail_out = zipInBufSize;
	switch ((ret = inflate(zin, Z_SYNC_FLUSH)) != 0) {
		case Z_OK:
			if (zin->avail_in != 0) {
				if (zin->avail_out != 0) {
					*len = -1;
					return zin->msg ? zin->msg : "????";
				}
				*nbuf = zin->next_in;
				*nbuflen = zin->avail_in;
			}
			*len = zipInBufSize - zin->avail_out;
			return zipInBuf;
		default:
			*len = -1;
			*err = ret;
			return zin->msg ? zin->msg : "????";
	}
}

int zip_is_data_out(void *session)
{
	ZLinkOut *z;
	if ((z = (ZLinkOut *)session) == NULL) {
		ircdlog(LOG_ERROR, "zip_is_data_out with NULL session pointer");
		return 0;
	}
	return z->bufsize;
}

char *zip_output(void *session, char *buffer, int *len, int forceflush, int *largedata)
{
	ZLinkOut *z;
	z_stream *zout;
	int ret;

	if ((z = (ZLinkOut *)session) == NULL) {
		ircdlog(LOG_ERROR, "zip_output() with NULL session pointer");
		*len = 0;
		return NULL;
	}
	zout = &z->stream;

	if (buffer != NULL) {
		memcpy(z->buf + z->bufsize, buffer, *len);
		z->bufsize += *len;
	}
	if (forceflush == 0 && (z->bufsize < ZIP_MIN_BLOCK || (largedata && z->bufsize < ZIP_MAX_BLOCK - 512))) {
		*len = 0;
		return NULL;
	}
	zout->next_in = z->buf;
	zout->avail_in = z->bufsize;
	zout->next_out = zipOutBuf;
	zout->avail_out = zipOutBufSize;
	if ((ret = deflate(zout, Z_SYNC_FLUSH)) == Z_OK) {
		z->bufsize = 0;
		*len = zipOutBufSize - zout->avail_out;
		return zipOutBuf;
	}
	*len = -1;
	*largedata = ret;
	return zout->msg ? zout->msg : "????";
}

void zip_out_get_stats(void *session, unsigned long *insiz, unsigned long *outsiz, double *ratio)
{
	ZLinkOut *z;
	if ((z = (ZLinkOut *)session) == NULL) {
		ircdlog(LOG_ERROR, "zip_out_get_stats() with NULL session pointer");
		*ratio = 0;
	}
	else {
		*outsiz = z->stream.total_out;
		if ((*insiz = z->stream.total_in)) {
			*ratio = (100.0 * (double)z->stream.total_out) / (double)z->stream.total_in;
		}
	}
	return;
}

void zip_destroy_output_session(void *session)
{
	ZLinkOut *z;
	if ((z = (ZLinkOut *)session) == NULL) {
		ircdlog(LOG_ERROR, "zip_destroy_output_session() with NULL session pointer");
	}
	else {
		deflateEnd(&z->stream);
		MyFree(session);
	}
	return;
}

void zip_destroy_input_session(void *session)
{
	ZLinkIn *z;
	if ((z = (ZLinkIn *)session) == NULL) {
		ircdlog(LOG_ERROR, "zip_destroy_input_session() with NULL session pointer");
	}
	else {
		inflateEnd(&z->stream);
		MyFree(session);
	}
	return;
}
